##
## A self-contained R-function to generate alternating codes
##
## (c) 2008 Juha Vierinen (j (at) sgo.fi)
##
## Usage:
## codes <- generateAlternatingCodes(minL=MIN_AC_LENGTH,maxL=MAX_AC_LENGTH)
##
generateAlternatingCodes <- function(minL=2,maxL=550)
{
  ss <- getSearchSpace(minL=minL,maxL=maxL)
  res <- walkSearchSpace(ss)
  return(res)
}

primes <- function(limit=1024)
{
  pr <- c()
  for(n in 1:limit)
  {
    if(is.prime(n))
      pr <- c(pr,n)
  }
  return(pr)
}

is.prime <- function(nn)
{
  res <- rep(F,length(nn))
  j <- 1
  for(n in nn)
  {
    if(n == 1 || n == 2)
    {
      res[j] <- T
    }
    else
    {
      res[j] <- T
      for(i in 2:(n-1))
      {
        if(divides(i,n))
        {
          res[j] <- F
          break
        }
      }
    }
    j <- j + 1
  }
  return(res)
}

cl2phmat <- function(cg)
{
  phmat <- array(dim=c(length(cg),length(cg[[1]])))
  for(ci in 1:length(cg))
  {
    phmat[ci,] <- cg[[ci]]
  }
  return(phmat)
}

phmat2cl <- function(phmat)
{
  cl <- list()
  for(i in 1:dim(phmat)[1])
  {
    cl[[i]] <- phmat[i,]
  }
  return(cl)
}

divides <- function(b,a)
{
  frac <- a/b
  if((floor(frac) - frac) == 0)
    return(T)
  return(F)
}

alternatingCodeLengths <- function(n,primelim=1024,explim=10)
{
  prs <- primes(primelim)
  prsm1 <- prs-1
  res <- c(prs,prsm1)
  for(i in 1:explim)
  {
    res <- c(res,prs**i)
  }
  return(unique(sort(res))[1:n])
}

getSearchSpace <- function(minL=2,maxL=128)
{
  prs <- primes(limit=maxL)
  ss <- array(dim=c(1,3))
  ss[1,] <- c(0,0,0)

  for(p in prs)
  {
    if(p > 1)
    {
      j <- 1
      while(p**j <= maxL)
      {
        if(p**j >= minL)
        {
          ss <- rbind(ss,c(p,j,p**j))
        }
        if(j == 1 && p >= 4)
        {
          ss <- rbind(ss,c(p-1,-1,p-1))
        }
        j <- j + 1
      }
    }
  }
  ss <- ss[2:dim(ss)[1],]
  ix <- sort(ss[,3],index.ret=T)$ix
  ss <- ss[ix,]
  return(ss)
}
nrepresentation <- function(n,base=2,length=10)
{
  rep <- vector(length=length)

  quotient <- n 
  for (i in 1:length)
  {
    rep[length-i+1] <- quotient %% base
    quotient <- quotient %/% base
  }
  rep
}
walkSearchSpace <- function(ss)
{
  res <- list()
  i <- 1
  for(si in 1:dim(ss)[1])
  {
    s <- ss[si,]
    p <- s[1]
    m <- s[2]
    if(m == -1)
    {
      ac <- findAC(p=p+1,m=1)
      ac <- createPM1(b=ac$b,p=p+1)
    }
    else
    {
      ac <- findAC(p=p,m=m)
    }
    sac <- createStrongCode(phmat2cl(t(ac$phases)))
    cat("m ",m,"b ",ac$b,"\n")
    res[[i]] <- list(p=p,m=m,l=s[3],ac=ac,sac=sac)
    i <- i + 1
  }
  return(res)
}
findAC <- function(p=2,m=1)
{
  b <- rep(0,m)
  maxn <- p**m
  for(i in 1:maxn)
  {
    bt <- nrepresentation(i,base=p,length=m)
    if(is.AC(b=bt,p=p))
      return(getAC(b=bt,p=p))
  }
}
differenceEq <- function(b=c(1,2,3),p=5,n=10,x=NULL)
{
  if(is.null(x))
  {
    x <- c(rep(0,length(b)-1),1)
  }

  for(i in (length(b)+1):n)
  {
    x[i] <- sum(b*x[(i-length(b)):(i-1)]) %% p
  }
  x
}
findPeriod <- function(x,maxp=5)
{
  mp=1
  for(i in 1:maxp)
  {
    test <- rep(x[1:i],2*floor(maxp/i))
    maxidx <- length(test)
    ss <- sum((test - x[1:maxidx])**2)
    if(ss == 0.0)
    {
      mp = i
      break
    }
  }
  return(mp)
}
is.AC <- function(b=c(2),p=3)
{
  maxper <- p**length(b)-1
  de <- differenceEq(b=b,p=p,n=(maxper*2+1))[2:(2*maxper+1)]
  per <- findPeriod(de,maxp=maxper)
  if(per == maxper)
    return(T)
  return(F)
}
createPM1 <- function(b=c(3),p=7)
{
  res <- list()
  res$alpha <- exp(2*pi*1i/(p-1))
  
  values <- differenceEq(b=b,p=p,n=(2*p))[1:(p-1)]
  indices <- 0:(length(values)-1)
  acpm1 <- rep(0,length(values))
  for(i in 1:length(values))
  {
    acpm1[values[i]] <- indices[i]
  }
  len <- p-1
  
  AC <- array(dim=c(len,len))
  AC[,] <- 0
  
  for(i in 1:len)
  {
    AC[i,] <- (acpm1*(i-1))%%len
  }
  res$powers <- t(AC)
  res$phases <- res$alpha**t(AC)
  res$b <- b
  res$n <- 1
  return(res)
}

#
# return alternating codes as phases and numbers
#
getAC <- function(b=c(2),p=5,len=(p**(length(b))-1))
{
  res <- list()
  res$alpha <- exp(2*pi*1i/p)
  
  m <- length(b)
  if(!is.AC(b=b,p=p))
    cat("Not alternating code\n")
  
  de <- differenceEq(b=b,p=p,n=(2*(len+1)))[2:((2*(len+1)))]
  AC <- array(dim=c(len+1,len+1))
  AC[,] <- 0

  for(i in 1:len)
  {
    AC[i,] <- de[i:(i+len)]
  }
  res$powers <- t(AC)
  res$phases <- res$alpha**t(AC)
  res$b <- b
  cat("number of codes \n")
  res$n <- totient2(p**length(b)-1)/length(b)
  cat(res$n,"\n")
  return(res)
}

totient2 <- function(n)
{
  cat("n ",n,"\n")
  tot <- c(0, 1, 2, 2, 4, 2, 6, 4, 6, 4, 10, 4, 12, 6, 8, 8, 16, 6, 18, 
           8, 12, 10, 22, 8, 20, 12, 18, 12, 28, 8, 30, 16, 20, 16, 24, 
           12, 36, 18, 24, 16, 40, 12, 42, 20, 24, 22, 46, 16, 42, 20, 32, 
           24, 52, 18, 40, 24, 36, 28, 58, 16, 60, 30, 36, 32, 48, 20, 66, 
           32, 44, 24, 70, 24, 72, 36, 40, 36, 60, 24, 78, 32, 54, 40, 82, 
           24, 64, 42, 56, 40, 88, 24, 72, 44, 60, 46, 72, 32, 96, 42, 60, 
           40, 100, 32, 102, 48, 48, 52, 106, 36, 108, 40, 72, 48, 112, 
           36, 88, 56, 72, 58, 96, 32, 110, 60, 80, 60, 100, 36, 126, 64, 
           84, 48, 130, 40, 108, 66, 72, 64, 136, 44, 138, 48, 92, 70, 120, 
           48, 112, 72, 84, 72, 148, 40, 150, 72, 96, 60, 120, 48, 156, 
           78, 104, 64, 132, 54, 162, 80, 80, 82, 166, 48, 156, 64, 108, 
           84, 172, 56, 120, 80, 116, 88, 178, 48, 180, 72, 120, 88, 144, 
           60, 160, 92, 108, 72, 190, 64, 192, 96, 96, 84, 196, 60, 198, 
           80, 132, 100, 168, 64, 160, 102, 132, 96, 180, 48, 210, 104, 
           140, 106, 168, 72, 180, 108, 144, 80, 192, 72, 222, 96, 120, 
           112, 226, 72, 228, 88, 120, 112, 232, 72, 184, 116, 156, 96, 
           238, 64, 240, 110, 162, 120, 168, 80, 216, 120, 164, 100, 250, 
           72, 220, 126, 128, 128, 256, 84, 216, 96, 168, 130, 262, 80, 
           208, 108, 176, 132, 268, 72, 270, 128, 144, 136, 200, 88, 276, 
           138, 180, 96, 280, 92, 282, 140, 144, 120, 240, 96, 272, 112, 
           192, 144, 292, 84, 232, 144, 180, 148, 264, 80, 252, 150, 200, 
           144, 240, 96, 306, 120, 204, 120, 310, 96, 312, 156, 144, 156, 
           316, 104, 280, 128, 212, 132, 288, 108, 240, 162, 216, 160, 276, 
           80, 330, 164, 216, 166, 264, 96, 336, 156, 224, 128, 300, 108, 
           294, 168, 176, 172, 346, 112, 348, 120, 216, 160, 352, 116, 280, 
           176, 192, 178, 358, 96, 342, 180, 220, 144, 288, 120, 366, 176, 
           240, 144, 312, 120, 372, 160, 200, 184, 336, 108, 378, 144, 252, 
           190, 382, 128, 240, 192, 252, 192, 388, 96, 352, 168, 260, 196, 
           312, 120, 396, 198, 216, 160, 400, 132, 360, 200, 216, 168, 360, 
           128, 408, 160, 272, 204, 348, 132, 328, 192, 276, 180, 418, 96, 
           420, 210, 276, 208, 320, 140, 360, 212, 240, 168, 430, 144, 432, 
           180, 224, 216, 396, 144, 438, 160, 252, 192, 442, 144, 352, 222, 
           296, 192, 448, 120, 399, 224, 300, 226, 288, 144, 456, 228, 288, 
           176, 460, 120, 462, 224, 240, 232, 466, 144, 396, 184, 312, 232, 
           420, 156, 360, 192, 312, 238, 478, 128, 432, 240, 264, 220, 384, 
           162, 486, 240, 324, 168, 490, 160, 448, 216, 240, 240, 420, 164, 
           498, 200, 332, 250, 502, 144, 400, 220, 312, 252, 508, 128, 432, 
           256, 324, 256, 408, 168, 460, 216, 344, 192, 520, 168, 522, 260, 
           240, 262, 480, 160, 506, 208, 348, 216, 480, 176, 424, 264, 356, 
           268, 420, 144, 540, 270, 360, 256, 432, 144, 546, 272, 360, 200, 
           504, 176, 468, 276, 288, 276, 556, 180, 504, 192, 320, 280, 562, 
           184, 448, 282, 324, 280, 568, 144, 570, 240, 380, 240, 440, 192, 
           576, 272, 384, 224, 492, 192, 520, 288, 288, 292, 586, 168, 540, 
           232, 392, 288, 592, 180, 384, 296, 396, 264, 598, 160, 600, 252, 
           396, 300, 440, 200, 606, 288, 336, 240, 552, 192, 612, 306, 320, 
           240, 616, 204, 618, 240, 396, 310, 528, 192, 500, 312, 360, 312, 
           576, 144, 630, 312, 420, 316, 504, 208, 504, 280, 420, 256, 640, 
           212, 642, 264, 336, 288, 646, 216, 580, 240, 360, 324, 652, 216, 
           520, 320, 432, 276, 658, 160, 660, 330, 384, 328, 432, 216, 616, 
           332, 444, 264, 600, 192, 672, 336, 360, 312, 676, 224, 576, 256, 
           452, 300, 682, 216, 544, 294, 456, 336, 624, 176, 690, 344, 360, 
           346, 552, 224, 640, 348, 464, 240, 700, 216, 648, 320, 368, 352, 
           600, 232, 708, 280, 468, 352, 660, 192, 480, 356, 476, 358, 718, 
           192, 612, 342, 480, 360, 560, 220, 726, 288, 486, 288, 672, 240, 
           732, 366, 336, 352, 659, 240, 738, 288, 432, 312, 742, 240, 592, 
           372, 492, 320, 636, 200, 750, 368, 500, 336, 600, 216, 756, 378, 
           440, 288, 760, 252, 648, 380, 384, 382, 696, 256, 768, 240, 512, 
           384, 772, 252, 600, 384, 432, 388, 720, 192, 700, 352, 504, 336, 
           624, 260, 786, 392, 524, 312, 672, 240, 720, 396, 416, 396, 796, 
           216, 736, 320, 528, 400, 719, 264, 528, 360, 536, 400, 808, 216, 
           810, 336, 540, 360, 648, 256, 756, 408, 432, 320, 820, 272, 822, 
           408, 400, 348, 826, 264, 828, 328, 552, 384, 672, 276, 664, 360, 
           540, 418, 838, 192, 812, 420, 560, 420, 624, 276, 660, 416, 564, 
           320, 792, 280, 852, 360, 432, 424, 856, 240, 858, 336, 480, 430, 
           862, 288, 688, 432, 544, 360, 780, 224, 792, 432, 576, 396, 600, 
           288, 876, 438, 584, 320, 880, 252, 882, 384, 464, 442, 886, 288, 
           756, 352, 540, 444, 828, 296, 712, 384, 528, 448, 840, 240, 832, 
           399, 504, 448, 720, 300, 906, 452, 600, 288, 910, 288, 820, 456, 
           480, 456, 780, 288, 918, 352, 612, 460, 840, 240, 720, 462, 612, 
           448, 928, 240, 756, 464, 620, 466, 640, 288, 936, 396, 624, 368, 
           940, 312, 880, 464, 432, 420, 946, 312, 864, 360, 632, 384, 952, 
           312, 760, 476, 560, 478, 816, 256, 930, 432, 636, 480, 768, 264, 
           966, 440, 576, 384, 970, 324, 828, 486, 480, 480, 976, 324, 880, 
           336, 648, 490, 982, 320, 784, 448, 552, 432, 924, 240, 990, 480, 
           660, 420, 792, 328, 996, 498, 648, 400)
  
  if(n > length(tot))
  {
    return(totient(n))
  }
  else
  {
    return(tot[n])
  }
}

printPhases <- function(phases)
{
  s <- sprintf("%d",phases[1])
  if(length(phases) > 1)
  {
    for(pi in 2:length(phases))
    {
      s <- sprintf("%s,%d",s,phases[pi])
    }
  }
  s
}

#
# 54 seems to fit in one latex page
printACResults <- function(res,foldLength=54)
{
  for(ri in 1:foldLength)
  {
    r <- res[[ri]]
    r2 <- res[[foldLength+ri]]
    r3 <- res[[2*foldLength+ri]]

    cat(sprintf("$%4d$ & $%4d$ & $%d$ & $",r$l,r$p,r$ac$n))
    if(r$m == -1)
    {
      cat(" \\bigstar $ & ")
    }
    else
    {
      cat(printPhases(r$ac$b))
      cat("$ & ")
    }

    cat(sprintf("$%4d$ & $%4d$ & $%d$ & $",r2$l,r2$p,r2$ac$n))
    if(r2$m == -1)
    {
      cat(" \\bigstar $ & ")
    }
    else
    {
      cat(printPhases(r2$ac$b))
      cat("$ & ")
    }
    cat(sprintf("$%4d$ & $%4d$ & $%d$ & $",r3$l,r3$p,r3$ac$n))
    if(r3$m == -1)
    {
      cat(" \\bigstar $  ")
    }
    else
    {
      cat(printPhases(r3$ac$b))
      cat("$ ")
    }
    cat("  \\\\ \n")
  }
}


plotLengths <- function(plott=F)
{
  a <- alternatingCodeLengths(1024)[3:1024]
  if(plott)
    pdf("lengths.pdf",width=5,height=5)
  plot(1:50,a[1:50],type="p",main="Alternating code lengths",ylab="Length",xlab="Code")

  p2 <- which(log2(a[1:50]) %% 1 == 0)
  points(p2,a[p2],col="black",pch=19)
  if(plott)
    dev.off()
}


cshift <- function(x,n=0)
{
  lx <- length(x)
  idx <- ((0+n):(lx-1+n)) %% lx + 1
  return(x[idx])
}

createGenWalsh <- function(p=6)
{
  W <- array(dim=c(p,p))
  for(i in 1:p)
  {
    for(j in 1:p)
    {
      W[i,] <- cshift(0:(p-1),n=(-i+1))
    }
  }
  return(W)
}

checkBFAC <- function(A)
{
  p <- dim(A)[1]
  
  # walk through all diagonals, stop if wrong
  for(i in 2:(p-1))
  {
    fidxs <- array(dim=c(2,(p-i+1)))
    fidxs[1,] <- i:p
    fidxs[2,] <- 1:(p-i+1)
    
    ll <- length(unique(A[t(fidxs)]))
    
    if(ll != (p-i+1))
      return(F)
  }
  return(T)
}

walkPermutations <- function(W)
{
  p <- dim(W)[1]
  perms <- permutations(p,p)
  mp <- dim(perms)[1]
  cat(sprintf("Walking throught %d perms\n",mp))
  res <- list()
  ri <- 1
  for(pi in 1:mp)
  {
#    cat(".")
    A <- W[perms[pi,],perms[pi,]]
    if(checkBFAC(A))
    {
      cat(sprintf("found %d\n",pi))
      res[[ri]] <- list(A=A,perm=perms[pi,])
      ri <- ri + 1
    }
  }
  return(res)
}

findBFAC <- function(p=6)
{
  W <- createGenWalsh(p=p)
  walkPermutations(W)
}

rperm <- function(len=6)
{
  p <- 1:len
  for(i in 1:(2*len))
  {
    i1 <- floor(runif(1)*len)+1
    i2 <- floor(runif(1)*len)+1
    while(i1 == i2)
      i2 <- floor(runif(1)*len)+1
    tmp <- p[i1]
    p[i1] <- p[i2]
    p[i2] <- tmp
  }
  return(p)
}

walkRPermutations <- function(W,n=1e6)
{
  res <- list()
  ri <- 1
  for(pi in 1:n)
  {
    perm <- rperm(len=dim(W)[1])
    A <- W[perm,perm]
    if(checkBFAC(A))
    {
      cat("found ",perm,"\n")
      res[[ri]] <- list(A=A,perm=perm)
      ri <- ri + 1
    }
  }
  return(res)
}

findRAC <- function(p=6,n=1e5)
{
  W <- createGenWalsh(p=p)
  res <- walkRPermutations(W,n=n)
  return(res)
}

# generate a strong code from a weak one
createStrongCode <- function(cg)
{
  scg <- list()
  mult <- (1:length(cg[[1]]) %% 2 - 1)
  mult[which(mult == 0)]  <- 1
  for(i in 1:length(cg))
  {
    scg[[i]] <- cg[[i]]
    scg[[i+length(cg)]] <- cg[[i]] * mult
  }
  return(scg)
}

testWeakCondition <- function(cg)
{
  N <- length(cg[[1]])
  K <- length(cg)

  s <- 0
  # for each i, i' j and j' where
  # j=i == j'-i'
  # i != j
  # i != i'
  for(k in 1:K)
  {
    for(i in 1:N)
    {
      for(ii in 1:N)
      {
        for(j in 1:N)
        {
          for(jj in 1:N)
          {
            if( (i-j) == (ii-jj) && i!=j && i != ii )
            {
              code <- cg[[k]]
              s <- s + code[i]*Conj(code[j])*Conj(code[ii]*Conj(code[jj]))
            }
          }
        }
      }
    }
  }
  small <- 1e-5
  if(abs(s) < small)
    return(T)
  else
    return(F)
}

testStrongCondition <- function(cg)
{
  N <- length(cg[[1]])
  K <- length(cg)

  s <- 0
  # for each i, i' j and j' where
  # j=i == j'-i'
  # i != j
  # i != i'
  for(k in 1:K)
  {
    for(i in 1:N)
    {
      for(ii in 1:N)
      {
        for(j in 1:N)
        {
          for(jj in 1:N)
          {
            if( ((i-j) == (ii-jj) || abs((i-j) - (ii-jj)) < 2 )&& i!=j && i != ii )
            {
              code <- cg[[k]]
              s <- s + code[i]*Conj(code[j])*Conj(code[ii]*Conj(code[jj]))
            }
          }
        }
      }
    }
  }
  small <- 1e-5
  if(abs(s) < small)
    return(T)
  else
    return(F)
}

checkACs <- function(res)
{
  for(r in res)
  {
#    v <- code.vars(t(r$ac$phases))
    v <- calcAll(phmat2cl(t(r$ac$phases)))
    cg <- phmat2cl(t(r$ac$phases))
    strong <- testStrongCondition(cg)
    weak <- testWeakCondition(cg)
    
    small <- 1e-6
    cat(sum(v$npwrs)/(r$l-1),"\n")
    if(abs(sum(v$npwrs)/(r$l-1) - 1) < small)
    {
      cat(sprintf("p %d m %d l %d ok strong %d weak %d\n",r$p,r$m,r$l,strong,weak))
    }
    else
    {
      cat(sprintf("p %d m %d l %d wrong\n",r$p,r$m,r$l))
      stop("not AC")
    }
      
  }
}

checkACs2 <- function(res)
{
  for(r in res)
  {
#    v <- code.vars(t(r$ac$phases))
#    v <- calcAll(phmat2cl(t(r$ac$phases)))
    cg <- phmat2cl(t(r$ac$phases))
    strong <- testStrongCondition(r$sac)
    weak <- testWeakCondition(cg)
    cat(sprintf("l %d p %d m %d ok strong %d weak %d\n",r$l,r$p,r$m,strong,weak))
  }
}

resultsToFile <- function(fname="results.R",res)
{
  ff <- file(fname,open="w")
  for(r in res)
  {
    cat(file=ff,sprintf("# Length %d \n",r$l))
    dput(r,control=NULL,file=ff)
    cat(file=ff,"\n\n")
  }
  close(f)
}


findPM1ACs <- function(p=6)
{
  perms <- permutations(p,p)
  res <- list()
  alpha <- exp(2*pi*1i/p)
  AC <- array(dim=c(p,p))
  AC[,] <- 0

  i<-1
  ri <- 1
  for(pi in 1:dim(perms)[1])
  {
    cat(sprintf("i %d/%d\n",i,dim(perms)[1]))
    A <- perms[pi,]-1
    cat(A,"\n")
    for(j in 1:p)
    {
      AC[j,] <- (A*j)%%p
    }
    powers <- t(AC)
    phases <- alpha**t(AC)
    if(testWeakCondition(phmat2cl(t(phases))))
    {
      return(A)
    }
    i <- i + 1
  }
  return(res)
}



doPaperTable <- function(maxL=550)
{
  ss <- getSearchSpace(minL=2,maxL=550)
  res <- walkSearchSpace(ss)
  printACResults(res,foldLength=54)
  return(res)
}

randomizeCodeGroup <- function(cg,p=30)
{
  rcg <- list()
  phases <- 2*pi*floor(runif(length(cg))*p)/p
  rvec <- exp(1i*phases)
  for(ri in 1:length(cg))
  {
    rcg[[ri]] <- rvec*cg[[ri]]
  }
  return(rcg)
}


findConjugate <- function(rcg)
{
  cglen <- length(rcg)
  cf <- T
  for( i in 1:cglen )
  {
    cfa <- F
    for( j in 1:cglen )
    {
      if( sum(abs(Conj(rcg[[j]]) - rcg[[i]])) < 1e-2 )
      {
        cfa <- T
        cat(i," - Conj(",j,") found\n")
      }
    }
    cf <- cf & cfa
  }
  return(cf)
}
  
